# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from enum import IntEnum
from collections.abc import Sized
import struct
from typing import (
    Iterable, List, Optional, Sequence, TypeVar, Union,
    cast, TYPE_CHECKING)

import numpy
from numpy import uint16, uint32

from spinn_utilities.overrides import overrides

from spinnman.model.enums import ExecutableType

from pacman.model.graphs import AbstractEdgePartition
from pacman.model.graphs.common import Slice
from pacman.model.graphs.machine import (
    MachineVertex, SourceSegmentedSDRAMMachinePartition, SDRAMMachineEdge)
from pacman.model.placements import Placement
from pacman.model.resources import AbstractSDRAM
from pacman.utilities.utility_calls import get_keys

from spinn_front_end_common.abstract_models import (
    AbstractHasAssociatedBinary,
    AbstractRewritesDataSpecification, AbstractGeneratesDataSpecification)
from spinn_front_end_common.interface.buffer_management import (
    recording_utilities)
from spinn_front_end_common.interface.buffer_management.buffer_models import (
    AbstractReceiveBuffersToHost)
from spinn_front_end_common.interface.ds import (
    DataType, DataSpecificationBase, DataSpecificationGenerator,
    DataSpecificationReloader)
from spinn_front_end_common.interface.profiling import (
    AbstractHasProfileData, ProfileData, profile_utils)
from spinn_front_end_common.interface.provenance import (
    ProvidesProvenanceDataFromMachineImpl)
from spinn_front_end_common.interface.simulation import simulation_utilities
from spinn_front_end_common.utilities import helpful_functions
from spinn_front_end_common.utilities.constants import (
    SIMULATION_N_BYTES, BYTES_PER_WORD, BYTES_PER_SHORT)
from spinn_front_end_common.utilities.helpful_functions import (
    locate_memory_region_for_placement)
from spinn_front_end_common.interface.provenance import ProvenanceWriter

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.exceptions import SynapticConfigurationException
from spynnaker.pyNN.models.abstract_models import (
    AbstractMaxSpikes, SendsSynapticInputsOverSDRAM,
    ReceivesSynapticInputsOverSDRAM)
from spynnaker.pyNN.utilities.constants import (
    LIVE_POISSON_CONTROL_PARTITION_ID)
from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
    NUMPY_CONNECTORS_DTYPE, ConnectionsArray)

if TYPE_CHECKING:
    from .spike_source_poisson_vertex import SpikeSourcePoissonVertex
    from spynnaker.pyNN.models.neural_projections import SynapseInformation
    from spynnaker.pyNN.models.neural_projections.connectors import (
        AbstractGenerateConnectorOnHost)

#: :meta private:
T = TypeVar("T")


def _flatten(alist: Iterable[Union[T, Iterable[T]]]) -> Iterable[T]:
    for item in alist:
        if hasattr(item, "__iter__"):
            yield from _flatten(item)
        else:
            yield item


def get_n_rates(vertex_slice: Slice, rate_data: Sequence[Sized]) -> int:
    """
    :returns: How many rates there are to be stored in total.
    """
    return sum(len(rate_data[i]) for i in vertex_slice.get_raster_ids())


def get_params_bytes(n_atoms: int) -> int:
    """
    :param n_atoms: How many atoms to account for
    :returns: The size of the Poisson parameters in bytes.
    """
    return (PARAMS_BASE_WORDS + n_atoms) * BYTES_PER_WORD


def get_rates_bytes(n_atoms: int, n_rates: int) -> int:
    """
    :param n_atoms: How many atoms to account for
    :param n_rates: How many rates to account for
    :returns: The size of the Poisson rates in bytes.
    """
    return ((n_atoms * PARAMS_WORDS_PER_NEURON) +
            (n_rates * PARAMS_WORDS_PER_RATE)) * BYTES_PER_WORD


def get_expander_rates_bytes(n_atoms: int, n_rates: int) -> int:
    """
    :param n_atoms: How many atoms to account for
    :param n_rates: How many rates to account for
    :returns: The size of the Poisson rates in bytes.
    """
    return ((n_atoms * EXPANDER_WORDS_PER_NEURON) +
            (n_rates * PARAMS_WORDS_PER_RATE) +
            EXPANDER_HEADER_WORDS) * BYTES_PER_WORD


def get_sdram_edge_params_bytes(vertex_slice: Slice) -> int:
    """
    :param vertex_slice:
    :returns: The size of the Poisson SDRAM region in bytes.
    """
    return SDRAM_EDGE_PARAMS_BASE_BYTES + (
        vertex_slice.n_atoms * SDRAM_EDGE_PARAMS_BYTES_PER_WEIGHT)


def _u3232_to_uint64(array: numpy.ndarray) -> numpy.ndarray:
    """
    Convert data to be written in U3232 to uint64 array.
    """
    return numpy.round(array * float(DataType.U3232.scale)).astype(
        DataType.U3232.numpy_typename)


# 1. uint32_t has_key;
# 2. uint32_t set_rate_neuron_id_mask;
# 3. UFRACT seconds_per_tick; 4. REAL ticks_per_second;
# 5. REAL slow_rate_per_tick_cut-off; 6. REAL fast_rate_per_tick_cut-off;
# 7. unt32_t first_source_id; 8. uint32_t n_spike_sources;
# 9. uint32_t max_spikes_per_timestep;
# 10. uint32_t n_colour_bits;
# 11,12,13,14 mars_kiss64_seed_t (uint[4]) spike_source_seed;
# 15. Rate changed flag
PARAMS_BASE_WORDS = 15

# uint32_t n_rates; uint32_t index
PARAMS_WORDS_PER_NEURON = 2

# unsigned long accum rate, start, duration
PARAMS_WORDS_PER_RATE = 6

# uint32_t count; one per neuron in worst case (= every neuron different)
EXPANDER_WORDS_PER_NEURON = PARAMS_WORDS_PER_NEURON + 1

# uint32_t n_items
EXPANDER_HEADER_WORDS = 2

# The size of each weight to be stored for SDRAM transfers
SDRAM_EDGE_PARAMS_BYTES_PER_WEIGHT = BYTES_PER_SHORT

# SDRAM edge param base size:
# 1. address, 2. size of transfer,
# 3. offset to start writing, 4. Array of weights (not counted here)
SDRAM_EDGE_PARAMS_BASE_BYTES = 3 * BYTES_PER_WORD

_ONE_WORD = struct.Struct("<I")
_FOUR_WORDS = struct.Struct("<4I")


class SpikeSourcePoissonMachineVertex(
        MachineVertex, AbstractReceiveBuffersToHost,
        ProvidesProvenanceDataFromMachineImpl,
        AbstractHasProfileData,
        AbstractHasAssociatedBinary, AbstractRewritesDataSpecification,
        AbstractGeneratesDataSpecification,
        SendsSynapticInputsOverSDRAM):
    """
    Vertex that implements a Poisson-distributed spike source.
    """

    __slots__ = (
        "__buffered_sdram_per_timestep",
        "__is_recording",
        "__minimum_buffer_sdram",
        "__sdram",
        "__sdram_partition",
        "__rate_changed")

    class _PoissonSpikeSourceRegions(IntEnum):
        """
        Memory region IDs for the the Poisson source code.
        """
        #: System control information (simulation timestep, etc.)
        SYSTEM_REGION = 0
        #: The parameters for the Poisson generator.
        POISSON_PARAMS_REGION = 1
        #: Spike rates (and the times at which they apply).
        RATES_REGION = 2
        #: Record of when spikes were actually sent.
        SPIKE_HISTORY_REGION = 3
        #: Provenance data.
        PROVENANCE_REGION = 4
        #: Profiler data.
        PROFILER_REGION = 5
        #: Parameters for an SDRAM edge.
        SDRAM_EDGE_PARAMS = 6
        #: Data for the on-chip connection generator binaries.
        EXPANDER_REGION = 7

    PROFILE_TAG_LABELS = {
        0: "TIMER",
        1: "PROB_FUNC"}

    # The maximum timestep - this is the maximum value of a uint32
    _MAX_TIMESTEP = 0xFFFFFFFF

    # between Exp and Knuth
    SLOW_RATE_PER_TICK_CUTOFF = 0.01

    # between Knuth algorithm and Gaussian approx.
    FAST_RATE_PER_TICK_CUTOFF = 10

    # Seed offset in parameters and size on bytes
    SEED_SIZE_BYTES = 4 * BYTES_PER_WORD
    SEED_OFFSET_BYTES = (PARAMS_BASE_WORDS * 4) - SEED_SIZE_BYTES

    def __init__(
            self, sdram: AbstractSDRAM, is_recording: bool,
            label: Optional[str], app_vertex: SpikeSourcePoissonVertex,
            vertex_slice: Slice):
        """
        :param sdram: SDRAM usage of this vertex
        :param is_recording: Flag to say spikes are to be recorded
        :param label: The optional name of the vertex
        :param app_vertex:
            The application vertex that caused this machine vertex to be
            created. If `None`, there is no such application vertex.
        :param vertex_slice:
            The slice of the application vertex that this machine vertex
            implements.
        """

        super().__init__(
            label, app_vertex=app_vertex, vertex_slice=vertex_slice)
        self.__is_recording = is_recording
        self.__sdram = sdram
        self.__sdram_partition: Optional[
            SourceSegmentedSDRAMMachinePartition] = None
        self.__rate_changed = True

    @property
    def _pop_vertex(self) -> SpikeSourcePoissonVertex:
        return cast('SpikeSourcePoissonVertex', self.app_vertex)

    @overrides(SendsSynapticInputsOverSDRAM.set_sdram_partition)
    def set_sdram_partition(
            self,
            sdram_partition: SourceSegmentedSDRAMMachinePartition) -> None:
        if self.__sdram_partition is not None:
            raise SynapticConfigurationException(
                "Cannot set SDRAM partition again")
        self.__sdram_partition = sdram_partition

    @property
    @overrides(MachineVertex.sdram_required)
    def sdram_required(self) -> AbstractSDRAM:
        return self.__sdram

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._provenance_region_id)
    def _provenance_region_id(self) -> int:
        return self._PoissonSpikeSourceRegions.PROVENANCE_REGION

    @overrides(ProvidesProvenanceDataFromMachineImpl
               .parse_extra_provenance_items)
    def parse_extra_provenance_items(
            self, label: str, x: int, y: int, p: int,
            provenance_data: Sequence[int]) -> None:
        (n_saturations,) = provenance_data
        with ProvenanceWriter() as db:
            db.insert_core(
                x, y, p, "Times_synaptic_weights_have_saturated",
                n_saturations)
            if n_saturations > 0:
                db.insert_report(
                    f"The weights from the Poisson synapses for {label}"
                    f" saturated {n_saturations} times. ")

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._n_additional_data_items)
    def _n_additional_data_items(self) -> int:
        return 1

    @overrides(AbstractReceiveBuffersToHost.get_recorded_region_ids)
    def get_recorded_region_ids(self) -> List[int]:
        if self.__is_recording:
            return [0]
        return []

    @overrides(AbstractReceiveBuffersToHost.get_recording_region_base_address)
    def get_recording_region_base_address(self, placement: Placement) -> int:
        return locate_memory_region_for_placement(
            placement, self._PoissonSpikeSourceRegions.SPIKE_HISTORY_REGION)

    @overrides(AbstractHasProfileData.get_profile_data)
    def get_profile_data(self, placement: Placement) -> ProfileData:
        return profile_utils.get_profiling_data(
            self._PoissonSpikeSourceRegions.PROFILER_REGION,
            self.PROFILE_TAG_LABELS, placement)

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self) -> str:
        return "spike_source_poisson.aplx"

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self) -> ExecutableType:
        return ExecutableType.USES_SIMULATION_INTERFACE

    @overrides(AbstractMaxSpikes.max_spikes_per_second)
    def max_spikes_per_second(self) -> float:
        # pylint: disable=missing-function-docstring
        # TODO https://github.com/SpiNNakerManchester/sPyNNaker/issues/1435
        return self._pop_vertex.max_rate

    @overrides(AbstractMaxSpikes.max_spikes_per_ts)
    def max_spikes_per_ts(self) -> float:
        # pylint: disable=missing-function-docstring
        # TODO https://github.com/SpiNNakerManchester/sPyNNaker/issues/1435
        return self._pop_vertex.max_spikes_per_ts()

    @overrides(AbstractRewritesDataSpecification.reload_required)
    def reload_required(self) -> bool:
        if self.__rate_changed:
            return True
        return SpynnakerDataView.get_first_machine_time_step() == 0

    @overrides(AbstractRewritesDataSpecification.set_reload_required)
    def set_reload_required(self, new_value: bool) -> None:
        self.__rate_changed = new_value

    @overrides(AbstractRewritesDataSpecification.regenerate_data_specification)
    def regenerate_data_specification(self, spec: DataSpecificationReloader,
                                      placement: Placement) -> None:
        # write rates
        self._write_poisson_rates(spec)

        # end spec
        spec.end_specification()

    @staticmethod
    def __conn(synapse_info: SynapseInformation
               ) -> AbstractGenerateConnectorOnHost:
        return cast(
            'AbstractGenerateConnectorOnHost', synapse_info.connector)

    @overrides(AbstractGeneratesDataSpecification.generate_data_specification)
    def generate_data_specification(self, spec: DataSpecificationGenerator,
                                    placement: Placement) -> None:
        spec.comment("\n*** Spec for SpikeSourcePoisson Instance ***\n\n")
        # if we are here, the rates have changed!
        self.__rate_changed = True

        # write setup data
        spec.reserve_memory_region(
            region=self._PoissonSpikeSourceRegions.SYSTEM_REGION,
            size=SIMULATION_N_BYTES, label='setup')
        spec.switch_write_focus(
            self._PoissonSpikeSourceRegions.SYSTEM_REGION)
        spec.write_array(simulation_utilities.get_simulation_header_array(
            self.get_binary_file_name()))

        # write recording data
        spec.reserve_memory_region(
            region=self._PoissonSpikeSourceRegions.SPIKE_HISTORY_REGION,
            size=recording_utilities.get_recording_header_size(1),
            label="Recording")
        spec.switch_write_focus(
            self._PoissonSpikeSourceRegions.SPIKE_HISTORY_REGION)
        sdram = self._pop_vertex.get_recording_sdram_usage(self.vertex_slice)
        recorded_region_sizes = [sdram.get_total_sdram(
            SpynnakerDataView.get_max_run_time_steps())]
        spec.write_array(recording_utilities.get_recording_header_array(
            recorded_region_sizes))

        # Write provenance space
        self.reserve_provenance_data_region(spec)

        # write parameters
        self._write_poisson_parameters(spec)

        # write rates
        self._write_poisson_rates(spec)

        # write profile data
        profile_utils.reserve_profile_region(
            spec, self._PoissonSpikeSourceRegions.PROFILER_REGION,
            self._pop_vertex.n_profile_samples)
        profile_utils.write_profile_region_data(
            spec, self._PoissonSpikeSourceRegions.PROFILER_REGION,
            self._pop_vertex.n_profile_samples)

        # write SDRAM edge parameters
        spec.reserve_memory_region(
            region=self._PoissonSpikeSourceRegions.SDRAM_EDGE_PARAMS,
            label="sdram edge params",
            size=get_sdram_edge_params_bytes(self.vertex_slice))
        spec.switch_write_focus(
            self._PoissonSpikeSourceRegions.SDRAM_EDGE_PARAMS)
        if self.__sdram_partition is None:
            spec.write_array([0, 0, 0])
        else:
            size = self.__sdram_partition.get_sdram_size_of_region_for(self)
            proj = self._pop_vertex.outgoing_projections[0]
            # pylint: disable=protected-access
            synapse_info = proj._synapse_information
            spec.write_value(
                self.__sdram_partition.get_sdram_base_address_for(self))
            spec.write_value(size)

            # Work out the offset into the data to write from, based on the
            # synapse type in use
            synapse_type = synapse_info.synapse_type
            offset = synapse_type * self.vertex_slice.n_atoms
            spec.write_value(offset)

            # If we are here, the connector must be one-to-one so create
            # the synapses and then store the scaled weights
            connections = self.__conn(synapse_info).create_synaptic_block(
                (), self.vertex_slice, synapse_type, synapse_info)
            weight_scales = (
                next(iter(cast(AbstractEdgePartition,
                               self.__sdram_partition).edges))
                .post_vertex.weight_scales)
            weights = connections["weight"] * weight_scales[synapse_type]
            weights = numpy.rint(numpy.abs(weights)).astype(uint16)
            if len(weights) % 2 != 0:
                padding = numpy.array([0], dtype=uint16)
                weights = numpy.concatenate((weights, padding))
            spec.write_array(weights.view(uint32))

        # End-of-Spec:
        spec.end_specification()

    def _write_poisson_rates(self, spec: DataSpecificationBase) -> None:
        """
        Generate Rate data for Poisson spike sources.

        :param spec: the data specification writer
        """
        spec.comment(
            f"\nWriting Rates for {self.vertex_slice.n_atoms} "
            "poisson sources:\n")

        n_atoms = self.vertex_slice.n_atoms
        n_rates = n_atoms * self._pop_vertex.max_n_rates
        spec.reserve_memory_region(
            region=self._PoissonSpikeSourceRegions.RATES_REGION,
            size=get_rates_bytes(n_atoms, n_rates), label='PoissonRates')

        # List starts with n_items, so start with 0.  Use arrays to allow
        # numpy concatenation to work.
        data_items: List[Union[Sequence[int], numpy.ndarray]] = list()
        data_items.append([int(self.__rate_changed)])
        data_items.append([0])
        n_items = 0
        data = self._pop_vertex.data
        ids = self.vertex_slice.get_raster_ids()
        for (start, stop, item) in data.iter_ranges_by_ids(ids):
            count = stop - start
            items = numpy.dstack(
                (_u3232_to_uint64(item['rates']),
                 _u3232_to_uint64(item['starts']),
                 _u3232_to_uint64(item['durations']))
                )[0]
            data_items.extend([[count], [len(items)], [0],
                               numpy.ravel(items).view(uint32)])
            n_items += 1
        data_items[1] = [n_items]
        data_to_write = numpy.concatenate(data_items)
        spec.reserve_memory_region(
            region=self._PoissonSpikeSourceRegions.EXPANDER_REGION,
            size=get_expander_rates_bytes(n_atoms, n_rates), label='Expander')
        spec.switch_write_focus(
            self._PoissonSpikeSourceRegions.EXPANDER_REGION)
        spec.write_array(data_to_write)

        self.__rate_changed = False

    def _write_poisson_parameters(self, spec: DataSpecificationBase) -> None:
        """
        Generate Parameter data for Poisson spike sources.

        :param spec: the data specification writer
        """
        spec.comment(
            f"\nWriting parameters for {self.vertex_slice.n_atoms} "
            "Poisson sources:\n")

        spec.reserve_memory_region(
            region=self._PoissonSpikeSourceRegions.POISSON_PARAMS_REGION,
            size=get_params_bytes(self.vertex_slice.n_atoms),
            label="PoissonParams")
        spec.switch_write_focus(
            self._PoissonSpikeSourceRegions.POISSON_PARAMS_REGION)

        # Write Key info for this core:
        routing_info = SpynnakerDataView.get_routing_infos()
        key = routing_info.get_single_key_from(self)
        keys: Union[Sequence[int], numpy.ndarray]
        if key is None:
            spec.write_value(0)
            keys = [0] * self.vertex_slice.n_atoms
        else:
            spec.write_value(1)
            keys = get_keys(
                key, self.vertex_slice, self._pop_vertex.n_colour_bits)
        # Write the incoming mask if there is one
        incoming_mask = 0
        if self._pop_vertex.incoming_control_edge is not None:
            routing_info = SpynnakerDataView.get_routing_infos()
            r_info = routing_info.get_info_from(
                self._pop_vertex.incoming_control_edge.pre_vertex,
                LIVE_POISSON_CONTROL_PARTITION_ID)
            incoming_mask = ~r_info.mask & 0xFFFFFFFF
        spec.write_value(incoming_mask)

        # Write the number of seconds per timestep (unsigned long fract)
        spec.write_value(
            data=SpynnakerDataView.get_simulation_time_step_s(),
            data_type=DataType.U032)

        # Write the number of timesteps per ms (accum)
        spec.write_value(
            data=SpynnakerDataView.get_simulation_time_step_per_ms(),
            data_type=DataType.U1616)

        # Write the slow rate per tick cut-off (accum)
        spec.write_value(
            data=self.SLOW_RATE_PER_TICK_CUTOFF, data_type=DataType.U1616)

        # Write the fast rate per tick cut-off (accum)
        spec.write_value(
            data=self.FAST_RATE_PER_TICK_CUTOFF, data_type=DataType.U1616)

        # Write the lo_atom ID
        spec.write_value(data=self.vertex_slice.lo_atom)

        # Write the number of sources
        spec.write_value(data=self.vertex_slice.n_atoms)

        # Write the maximum spikes per tick
        spec.write_value(data=self.max_spikes_per_ts())

        # Write the number of colour bits
        spec.write_value(data=self._pop_vertex.n_colour_bits)

        # Write the random seed (4 words), generated randomly!
        spec.write_array(self._pop_vertex.kiss_seed(self.vertex_slice))

        spec.write_array(keys)

    def set_rate_changed(self) -> None:
        """
        Records that the rates have changed.
        """
        self.__rate_changed = True

    def __poisson_rate_region_address(self, placement: Placement) -> int:
        return helpful_functions.locate_memory_region_for_placement(
            placement, self._PoissonSpikeSourceRegions.RATES_REGION)

    def read_parameters_from_machine(self, placement: Placement) -> None:
        """
        Reads the poisson rates of the machine if they could have changed.

        :param placement:
        """
        # It is only worth updating the rates when there is a control edge
        # that can change them
        if self._pop_vertex.incoming_control_edge is not None:
            # locate SDRAM address where the rates are stored
            poisson_rate_region_sdram_address = (
                self.__poisson_rate_region_address(placement))

            # get size of poisson params
            n_atoms = self.vertex_slice.n_atoms
            n_rates = n_atoms * self._pop_vertex.max_n_rates
            size_of_region = get_rates_bytes(n_atoms, n_rates)

            # get data from the machine
            byte_array = SpynnakerDataView.read_memory(
                placement.x, placement.y,
                poisson_rate_region_sdram_address, size_of_region)

            # For each atom, read the number of rates and the rate parameters
            offset = 0
            for i in self.vertex_slice.get_raster_ids():
                n_rates, = _ONE_WORD.unpack_from(byte_array, offset)
                # Skip the count and index
                offset += PARAMS_WORDS_PER_NEURON * BYTES_PER_WORD
                rates: List[float] = list()
                for _ in range(n_rates):
                    rate_int = _ONE_WORD.unpack_from(byte_array, offset)[0]
                    rates.append(rate_int / DataType.S1615.scale)
                    # Skip the start and duration as they can't change
                    offset += PARAMS_WORDS_PER_RATE * BYTES_PER_WORD
                self._pop_vertex.rates.set_value_by_id(i, numpy.array(rates))

    def read_connections(
            self, synapse_info: SynapseInformation) -> ConnectionsArray:
        """
        Read the connections from the machine.

        :param synapse_info: The synapse information being read
        :return: The connections read back
        """
        size = self.vertex_slice.n_atoms * SDRAM_EDGE_PARAMS_BYTES_PER_WEIGHT
        placement = SpynnakerDataView().get_placement_of_vertex(self)
        addr = locate_memory_region_for_placement(
            placement, self._PoissonSpikeSourceRegions.SDRAM_EDGE_PARAMS)
        addr += SDRAM_EDGE_PARAMS_BASE_BYTES
        data = SpynnakerDataView().read_memory(
            placement.x, placement.y, addr, size)
        weights_encoded = numpy.frombuffer(data, dtype=uint16).astype(
            numpy.float64)
        weight_scales = (
                next(iter(cast(AbstractEdgePartition,
                               self.__sdram_partition).edges))
                .post_vertex.weight_scales)
        weights = weights_encoded / weight_scales[synapse_info.synapse_type]
        connections = numpy.zeros(
            self.vertex_slice.n_atoms, dtype=NUMPY_CONNECTORS_DTYPE)
        connections["source"] = numpy.arange(
            self.vertex_slice.lo_atom, self.vertex_slice.hi_atom + 1)
        connections["target"] = numpy.arange(
            self.vertex_slice.lo_atom, self.vertex_slice.hi_atom + 1)
        connections["weight"] = weights
        connections["delay"] = \
            SpynnakerDataView().get_simulation_time_step_ms()
        return connections

    @overrides(SendsSynapticInputsOverSDRAM.sdram_requirement)
    def sdram_requirement(self, sdram_machine_edge: SDRAMMachineEdge) -> int:
        if isinstance(sdram_machine_edge.post_vertex,
                      ReceivesSynapticInputsOverSDRAM):
            return sdram_machine_edge.post_vertex.n_bytes_for_transfer
        raise SynapticConfigurationException(
            f"Unknown post vertex type in edge {sdram_machine_edge}")

    @overrides(MachineVertex.get_n_keys_for_partition)
    def get_n_keys_for_partition(self, partition_id: str) -> int:
        return self.vertex_slice.n_atoms << self._pop_vertex.n_colour_bits



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Iterator, MutableMapping, Optional, Tuple, cast
from typing_extensions import Never

import numpy
from numpy import uint32
from numpy.typing import NDArray

from spinn_utilities.helpful_functions import is_singleton
from spinn_utilities.ranged import RangeDictionary, RangedList

from pacman.model.graphs.common import Slice
from pacman.model.placements import Placement

from spinn_front_end_common.interface.ds import DataSpecificationBase
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD
from spinn_front_end_common.utilities.helpful_functions import (
    locate_memory_region_for_placement)

from spynnaker.pyNN.utilities.struct import Struct, StructRepeat
from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.models.neuron.population_vertex import (
    PopulationVertex)
from spynnaker.pyNN.models.neuron.population_machine_neurons import (
    NeuronRegions)


def _all_one_val_gen(rd: RangeDictionary[float]) -> bool:
    """
    Determine if all the values of a dictionary are the same, assuming we
    already know that they are generatable

    .. note::
        A random distribution is considered the same if the same distribution
        is used for all neurons.
    """
    for key in rd.keys():
        if is_singleton(rd[key]):
            return True
        if not rd[key].range_based():
            return False
        for i, (_start, _stop, _val) in enumerate(rd[key].iter_ranges()):
            if i > 0:
                return False
    return True


class NeuronData(object):
    """
    Holds and creates the data for a group of neurons.
    """

    __slots__ = (
        # The application vertex
        "__app_vertex",

        # The data to be written for all vertices, if applicable
        "__neuron_data",

        # The number of structs in neuron_data
        "__neuron_data_n_structs",

        # The neuron recording data for all vertices, if applicable
        "__neuron_recording_data",

        # Whether an attempt has been made to generate neuron data
        "__generation_done",

        # Whether to generate things on the machine
        "__gen_on_machine")

    def __init__(self, app_vertex: PopulationVertex):
        """
        :param app_vertex: Vertex to get data from.
        """
        self.__app_vertex = app_vertex
        self.__neuron_data: Optional[NDArray[uint32]] = None
        self.__neuron_recording_data: Optional[NDArray[uint32]] = None
        self.__generation_done = False
        self.__gen_on_machine = False
        self.__neuron_data_n_structs = 0

    @property
    def gen_on_machine(self) -> bool:
        """
        Whether the neuron data can be generated on the machine or not.
        """
        return self.__gen_on_machine

    def generate_data(self) -> None:
        """
        Do the data generation internally.
        """
        if self.__generation_done:
            return
        self.__generation_done = True

        if not self.__app_vertex.can_generate_on_machine():
            self.__gen_on_machine = False
            return

        # Check that all parameters and state variables have a single range
        params = self.__app_vertex.parameters
        state_vars = self.__app_vertex.state_variables
        if not _all_one_val_gen(params) or not _all_one_val_gen(state_vars):
            # Note at this point, we can still generate ranges on machine,
            # just that it has to be different per core
            self.__gen_on_machine = True
            self.__neuron_data = None
            return

        # Go through all the structs and make all the data
        structs = self.__app_vertex.neuron_impl.structs
        values = _MergedDict(params, state_vars)
        all_data = [struct.get_generator_data(values) for struct in structs]
        self.__neuron_data = numpy.concatenate(all_data)
        self.__neuron_data_n_structs = len(structs)

        # Get the neuron recording data
        neuron_recorder = self.__app_vertex.neuron_recorder
        if neuron_recorder.is_global_generatable:
            self.__neuron_recording_data = neuron_recorder.get_generator_data()

        # If we get here, we know everything is generated on machine
        self.__gen_on_machine = True

    def write_data(self, spec: DataSpecificationBase, vertex_slice: Slice,
                   neuron_regions: NeuronRegions,
                   gen_on_machine: bool = True) -> None:
        """
        Write the generated data.

        :param spec: The data specification to write to
        :param vertex_slice: The vertex slice to generate for
        :param neuron_regions: The regions to write to
        :param gen_on_machine: Whether to allow generation on machine
        """
        if gen_on_machine:
            self.generate_data()
        spec.reserve_memory_region(
            region=neuron_regions.neuron_params,
            size=self.__app_vertex.get_sdram_usage_for_neuron_params(
                    vertex_slice.n_atoms),
            label="neuron_params")
        neuron_recorder = self.__app_vertex.neuron_recorder
        spec.reserve_memory_region(
            region=neuron_regions.neuron_recording,
            size=neuron_recorder.get_metadata_sdram_usage_in_bytes(
                vertex_slice.n_atoms),
            label="neuron recording")
        if self.gen_on_machine and gen_on_machine:
            if self.__neuron_data is not None:
                data = self.__neuron_data
                n_structs = self.__neuron_data_n_structs
            else:
                n_structs, data = self.__get_neuron_builder_data(vertex_slice)
            header = self.__get_neuron_builder_header(
                vertex_slice, n_structs, neuron_regions)
            if self.__neuron_recording_data is not None:
                rec_data = self.__neuron_recording_data
            else:
                rec_data = neuron_recorder.get_generator_data(vertex_slice)
            n_words = len(data) + len(header) + len(rec_data)
            spec.reserve_memory_region(
                region=neuron_regions.neuron_builder,
                size=n_words * BYTES_PER_WORD,
                label="neuron_builder")
            spec.switch_write_focus(neuron_regions.neuron_builder)
            spec.write_array(header)
            spec.write_array(data)
            spec.write_array(rec_data)
        else:
            spec.switch_write_focus(neuron_regions.neuron_params)
            neuron_data = self.__get_neuron_param_data(vertex_slice)
            spec.write_array(neuron_data)
            neuron_recorder.write_neuron_recording_region(
                spec, neuron_regions.neuron_recording, vertex_slice)
        spec.reserve_memory_region(
            region=neuron_regions.initial_values,
            size=self.__app_vertex.get_sdram_usage_for_neuron_params(
                    vertex_slice.n_atoms),
            label="initial_values")

    def __get_neuron_param_data(self, vertex_slice: Slice) -> NDArray:
        """
        Get neuron parameter data for a slice.

        :param vertex_slice: The slice to get the data for
        """
        structs = self.__app_vertex.neuron_impl.structs
        values = _MergedDict(self.__app_vertex.parameters,
                             self.__app_vertex.state_variables)
        return numpy.concatenate([
            self.__get_struct_data(struct, values, vertex_slice)
            for struct in structs])

    def __get_struct_data(self, struct: Struct, values: '_MergedDict',
                          vertex_slice: Slice) -> NDArray[uint32]:
        """
        Get the data for a structure.

        :param struct: The structure to get the data for
        :param values: The values to fill in the structure with
        :param vertex_slice: The slice to get the values for
        """
        if struct.repeat_type == StructRepeat.GLOBAL:
            return struct.get_data(values)
        return struct.get_data(values, vertex_slice)

    def __get_neuron_builder_data(
            self, vertex_slice: Slice) -> Tuple[int, NDArray[uint32]]:
        """
        Get the data to build neuron parameters with.

        :param vertex_slice: The slice to get the parameters for
        :return: The number of structures and the data
        """
        structs = self.__app_vertex.neuron_impl.structs
        values = _MergedDict(self.__app_vertex.parameters,
                             self.__app_vertex.state_variables)
        return len(structs), numpy.concatenate([
            self.__get_builder_data(struct, values, vertex_slice)
            for struct in structs])

    def __get_builder_data(
            self, struct: Struct, values: '_MergedDict',
            vertex_slice: Slice) -> NDArray[uint32]:
        """
        Get the builder data for a structure.

        :param struct: The structure to get the data for
        :param values: The values to fill in the structure with
        :param vertex_slice: The slice to get the values for
        """
        if struct.repeat_type == StructRepeat.GLOBAL:
            return struct.get_generator_data(values)
        return struct.get_generator_data(values, vertex_slice)

    def __get_neuron_builder_header(
            self, vertex_slice: Slice, n_structs: int,
            neuron_regions: NeuronRegions) -> NDArray[uint32]:
        """
        Get the header of the neuron builder region.

        :param vertex_slice: The slice to put in the header
        :param n_structs: The number of structures to generate
        :param neuron_regions: The regions to point to
        """
        return numpy.array([
            neuron_regions.neuron_params,
            neuron_regions.neuron_recording,
            *self.__app_vertex.pop_seed,
            *self.__app_vertex.core_seed(vertex_slice),
            n_structs, vertex_slice.n_atoms], dtype=uint32)

    def read_data(self, placement: Placement,
                  neuron_regions: NeuronRegions) -> None:
        """
        Read the current state of the data from the machine into the
        application vertex.

        :param placement:
            The placement of the vertex to read
        :param neuron_regions: The regions to read from
        """
        merged_dict = _MergedDict(self.__app_vertex.parameters,
                                  self.__app_vertex.state_variables)
        self.__do_read_data(
            placement, neuron_regions.neuron_params, merged_dict)

    def read_initial_data(self, placement: Placement,
                          neuron_regions: NeuronRegions) -> None:
        """
        Read the initial state of the data from the machine into the
        application vertex.

        :param placement:
            The placement of the vertex to read
        :param neuron_regions: The regions to read from
        """
        merged_dict = _MergedDict(self.__app_vertex.parameters,
                                  self.__app_vertex.initial_state_variables)
        self.__do_read_data(
            placement, neuron_regions.initial_values, merged_dict)

    def __do_read_data(self, placement: Placement, region: int,
                       results: '_MergedDict') -> None:
        """
        Perform the reading of data.

        :param placement: Where the vertex is on the machine
        :param region: The region to read from
        :param results: Where to write the results to
        """
        address = locate_memory_region_for_placement(placement, region)
        vertex_slice = placement.vertex.vertex_slice
        data_size = self.__app_vertex.get_sdram_usage_for_neuron_params(
            vertex_slice.n_atoms)
        block = SpynnakerDataView.read_memory(
            placement.x, placement.y, address, data_size)
        offset = 0
        for struct in self.__app_vertex.neuron_impl.structs:
            if struct.repeat_type == StructRepeat.GLOBAL:
                struct.read_data(block, cast(RangeDictionary, results), offset)
                offset += struct.get_size_in_whole_words() * BYTES_PER_WORD
            else:
                struct.read_data(block, cast(RangeDictionary, results), offset,
                                 vertex_slice)
                offset += (
                    struct.get_size_in_whole_words(vertex_slice.n_atoms) *
                    BYTES_PER_WORD)

    def reset_generation(self) -> None:
        """
        Reset generation so it is done again.
        """
        self.__neuron_data = None
        self.__neuron_recording_data = None
        self.__generation_done = False
        self.__gen_on_machine = False
        self.__neuron_data_n_structs = 0


class _MergedDict(MutableMapping[str, RangedList[float]]):
    __slots__ = (
        "__params",
        "__state_vars")

    def __init__(self, params: RangeDictionary[float],
                 state_vars: RangeDictionary[float]):
        """
        :param params: The parameter values
        :param state_vars: The State values
        """
        self.__params = params
        self.__state_vars = state_vars

    def __contains__(self, key: object) -> bool:
        if not isinstance(key, str):
            return False
        return key in self.__params or key in self.__state_vars

    def __getitem__(self, key: str) -> RangedList[float]:
        if key in self.__params:
            return self.__params[key]
        return self.__state_vars[key]

    def __setitem__(self, key: str, value: RangedList[float]) -> None:
        if key in self.__params:
            self.__params[key] = value
        elif key in self.__state_vars:
            self.__state_vars[key] = value
        else:
            raise KeyError(f"No such key {key}")

    def __delitem__(self, __v: str) -> Never:
        raise NotImplementedError("items may not be deleted")

    def __iter__(self) -> Iterator[str]:
        yield from self.__params.keys()
        yield from self.__state_vars.keys()

    def __len__(self) -> int:
        return len(self.__params) + len(self.__state_vars)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable

from numpy import floating
from numpy.typing import NDArray

from spinn_utilities.overrides import overrides

from spinn_front_end_common.interface.ds import DataSpecificationBase
from spinn_front_end_common.utilities.constants import (
    BYTES_PER_SHORT, BYTES_PER_WORD)

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.models.neuron.plasticity.stdp.common import (
    get_exp_lut_array)
from spynnaker.pyNN.models.neuron.plasticity.stdp.timing_dependence import (
    AbstractTimingDependence)
from spynnaker.pyNN.models.neuron.plasticity.stdp.synapse_structure import (
    SynapseStructureWeightOnly)


class TimingDependencePfisterSpikeTriplet(AbstractTimingDependence):
    # pylint: disable=wrong-spelling-in-docstring
    """
    A timing dependence STDP rule based on spike triplets.

    Jean-Pascal Pfister, Wulfram Gerstner. Triplets of Spikes in a Model of
    Spike Timing-Dependent Plasticity. *Journal of Neuroscience*,
    20 September 2006, 26 (38) 9673-9682; DOI: 10.1523/JNEUROSCI.1425-06.2006
    """
    __slots__ = (
        "__tau_minus",
        "__tau_minus_data",
        "__tau_plus",
        "__tau_plus_data",
        "__tau_x",
        "__tau_x_data",
        "__tau_y",
        "__tau_y_data",
        "__a_plus",
        "__a_minus")
    __PARAM_NAMES = ('tau_plus', 'tau_minus', 'tau_x', 'tau_y')

    # noinspection PyPep8Naming
    def __init__(
            self, tau_plus: float, tau_minus: float,
            tau_x: float, tau_y: float, A_plus: float, A_minus: float):
        r"""
        :param tau_plus: :math:`\tau_+`
        :param tau_minus: :math:`\tau_-`
        :param tau_x: :math:`\tau_x`
        :param tau_y: :math:`\tau_y`
        :param A_plus: :math:`A^+`
        :param A_minus: :math:`A^-`
        """
        super().__init__(SynapseStructureWeightOnly())
        self.__tau_plus = tau_plus
        self.__tau_minus = tau_minus
        self.__tau_x = tau_x
        self.__tau_y = tau_y
        self.__a_plus = A_plus
        self.__a_minus = A_minus

        ts = SpynnakerDataView.get_simulation_time_step_ms()
        self.__tau_plus_data = get_exp_lut_array(ts, self.__tau_plus)
        self.__tau_minus_data = get_exp_lut_array(ts, self.__tau_minus)
        self.__tau_x_data = get_exp_lut_array(ts, self.__tau_x, shift=2)
        self.__tau_y_data = get_exp_lut_array(ts, self.__tau_y, shift=2)

    @property
    def tau_plus(self) -> float:
        r"""
        :math:`\tau_+`
        """
        return self.__tau_plus

    @property
    def tau_minus(self) -> float:
        r"""
        :math:`\tau_-`
        """
        return self.__tau_minus

    @property
    def tau_x(self) -> float:
        r"""
        :math:`\tau_x`
        """
        return self.__tau_x

    @property
    def tau_y(self) -> float:
        r"""
        :math:`\tau_y`
        """
        return self.__tau_y

    @property
    def A_plus(self) -> float:
        r"""
        :math:`A^+`
        """
        return self.__a_plus

    @A_plus.setter
    def A_plus(self, new_value: float) -> None:
        self.__a_plus = new_value

    @property
    def A_minus(self) -> float:
        r"""
        :math:`A^-`
        """
        return self.__a_minus

    @A_minus.setter
    def A_minus(self, new_value: float) -> None:
        self.__a_minus = new_value

    @overrides(AbstractTimingDependence.is_same_as)
    def is_same_as(self, timing_dependence: AbstractTimingDependence) -> bool:
        if not isinstance(
                timing_dependence, TimingDependencePfisterSpikeTriplet):
            return False
        return (
            (self.__tau_plus == timing_dependence.tau_plus) and
            (self.__tau_minus == timing_dependence.tau_minus) and
            (self.__tau_x == timing_dependence.tau_x) and
            (self.__tau_y == timing_dependence.tau_y))

    @property
    def vertex_executable_suffix(self) -> str:
        """
        The suffix to be appended to the vertex executable for this rule.
        """
        return "pfister_triplet"

    @property
    def pre_trace_n_bytes(self) -> int:
        """
        The number of bytes used by the pre-trace of the rule per neuron.
        """
        # Triplet rule trace entries consists of two 16-bit traces - R1 and R2
        # (Note: this is the pre-trace size, not the post-trace size)
        return BYTES_PER_SHORT * 2

    @overrides(AbstractTimingDependence.get_parameters_sdram_usage_in_bytes)
    def get_parameters_sdram_usage_in_bytes(self) -> int:
        lut_array_words = (
            len(self.__tau_plus_data) + len(self.__tau_minus_data) +
            len(self.__tau_x_data) + len(self.__tau_y_data))
        return lut_array_words * BYTES_PER_WORD

    @property
    def n_weight_terms(self) -> int:
        """
        The number of weight terms expected by this timing rule.
        """
        return 2

    @overrides(AbstractTimingDependence.write_parameters)
    def write_parameters(
            self, spec: DataSpecificationBase, global_weight_scale: float,
            synapse_weight_scales: NDArray[floating]) -> None:
        # Write lookup tables
        spec.write_array(self.__tau_plus_data)
        spec.write_array(self.__tau_minus_data)
        spec.write_array(self.__tau_x_data)
        spec.write_array(self.__tau_y_data)

    @overrides(AbstractTimingDependence.get_parameter_names)
    def get_parameter_names(self) -> Iterable[str]:
        return self.__PARAM_NAMES



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, Iterable, List, Optional, Tuple

from spinn_utilities.overrides import overrides

from pacman.model.graphs.machine import MachineFPGAVertex
from pacman.model.graphs.application import (
    Application2DFPGAVertex, FPGAConnection)
from pacman.model.graphs.common import Slice
from pacman.model.graphs.machine import MachineVertex
from pacman.model.routing_info import BaseKeyAndMask, RoutingInfo
from pacman.utilities.constants import BITS_IN_KEY
from pacman.utilities.utility_calls import is_power_of_2

from spinn_front_end_common.abstract_models import (
    AbstractSendMeMulticastCommandsVertex)
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utility_models import MultiCastCommand

from spynnaker.pyNN.models.common import PopulationApplicationVertex
from .spif_devices import (
    SPIF_FPGA_ID, SPIF_OUTPUT_FPGA_LINK, SPIF_INPUT_FPGA_LINKS,
    N_PIPES, N_FILTERS, SpiNNFPGARegister, SPIFRegister,
    set_field_mask, set_field_shift, set_field_limit,
    set_filter_mask, set_filter_value, set_mapper_key,
    set_input_key, set_input_mask, set_input_route)


class SPIFRetinaDevice(
        Application2DFPGAVertex, PopulationApplicationVertex,
        AbstractSendMeMulticastCommandsVertex):
    """
    A retina device connected to SpiNNaker using a SPIF board.
    """

    #: SPIF outputs to 8 FPGA output links, so we split into (2 x 4), meaning
    #: a mask of (1 x 3)
    Y_MASK = 1

    #: See Y_MASK for description
    X_MASK = 3

    #: The number of X values per row
    X_PER_ROW = 4

    #: The number of devices in existence, to work out the key
    __n_devices = 0

    __slots__ = (
        "__spif_mask",
        "__index_by_slice",
        "__base_key",
        "__pipe",
        "__input_y_mask",
        "__input_y_shift",
        "__input_x_mask",
        "__input_x_shift")

    @classmethod
    def __issue_device_id(cls, base_key: Optional[int]) -> int:
        if base_key is None:
            base_key = cls.__n_devices
        cls.__n_devices += 1
        return base_key

    def __init__(self, pipe: int, width: int, height: int, sub_width: int,
                 sub_height: int, base_key: Optional[int] = None,
                 input_x_shift: int = 16, input_y_shift: int = 0,
                 board_address: Optional[str] = None,
                 chip_coords: Optional[Tuple[int, int]] = None):
        """
        :param pipe: Which pipe on SPIF the retina is connected to
        :param width: The width of the retina in pixels
        :param height: The height of the retina in pixels
        :param sub_width:
            The width of rectangles to split the retina into for efficiency of
            sending
        :param sub_height:
            The height of rectangles to split the retina into for efficiency of
            sending
        :param base_key:
            The key that is common over the whole vertex,
            or `None` to use the pipe number as the key
        :param input_x_shift:
            The shift to get the x coordinate from the input keys sent to SPIF
        :param input_y_shift:
            The shift to get the y coordinate from the input keys sent to SPIF
        :param board_address:
            The IP address of the board to which the FPGA is connected,
            or `None` to use the default board or chip_coords.

            .. note::
                chip_coords will be used first if both are specified, with
                board_address then being used if the coordinates don't connect
                to an FPGA.
        :param chip_coords:
            The coordinates of the chip to which the FPGA is connected, or
            `None` to use the default board or board_address.

            .. note::
                chip_coords will be used first if board_address is also
                specified, with board_address then being used if the
                coordinates don't connect to an FPGA.
        """
        # Do some checks
        if sub_width < self.X_MASK + 1 or sub_height < self.Y_MASK + 1:
            raise ConfigurationException(
                "The sub-squares must be >=4 x >= 2"
                f" ({sub_width} x {sub_height} specified)")
        if pipe >= N_PIPES:
            raise ConfigurationException(
                f"Pipe {pipe} is bigger than maximum allowed {N_PIPES}")

        # The width has to be a power of 2 as otherwise the keys will not line
        # up correctly (x is at the least significant bit of the key,
        # so key then has an x # field).
        # This is an error here as it affects the downstream
        # population calculations also!
        if not is_power_of_2(width):
            raise ConfigurationException(
                "The width of the SPIF retina must be a power of 2.  If the"
                " real retina size is less than this, please round it up."
                " This will ensure that following Populations can decode the"
                " spikes correctly.  Note that you will also have to make the"
                " sizes of the following Populations bigger to match!")

        # Call the super
        super().__init__(
            width, height, sub_width, sub_height,
            self.__incoming_fpgas(board_address, chip_coords),
            self.__outgoing_fpga(board_address, chip_coords))

        # The mask is going to be made up of:
        # | K | P | Y_I | Y_0 | Y_F | X_I | X_0 | X_F |
        # K = base key
        # P = polarity (0 as not cared about)
        # Y_I = y index of sub-square
        # Y_0 = 0s for values not cared about in Y
        # Y_F = FPGA y index
        # X_I = x index of sub-square
        # X_0 = 0s for values not cared about in X
        # X_F = FPGA x index
        # Now - go calculate:
        x_bits = self._x_bits
        y_bits = self._y_bits

        # Mask to apply to route packets at input
        n_key_bits = BITS_IN_KEY - self._key_shift
        key_mask = (1 << n_key_bits) - 1
        self.__spif_mask = (
            (key_mask << self._key_shift) +
            (self.Y_MASK << self._source_y_shift) +
            (self.X_MASK << self._source_x_shift))

        # A dictionary to get vertex index from FPGA and slice
        self.__index_by_slice: Dict[Tuple[int, Slice], int] = dict()

        self.__pipe = pipe
        self.__base_key = self.__issue_device_id(base_key)

        # Generate the shifts and masks to convert the SPIF Ethernet inputs to
        # PYX format
        self.__input_x_mask = ((1 << x_bits) - 1) << input_x_shift
        self.__input_x_shift = self.__unsigned(input_x_shift)
        self.__input_y_mask = ((1 << y_bits) - 1) << input_y_shift
        self.__input_y_shift = self.__unsigned(input_y_shift - x_bits)

    @staticmethod
    def __unsigned(n: int) -> int:
        return n & 0xFFFFFFFF

    def __incoming_fpgas(
            self, board_address: Optional[str],
            chip_coords: Optional[Tuple[int, int]]) -> List[FPGAConnection]:

        """
        Get the incoming FPGA connections.
        """
        # We use every other odd link
        return [FPGAConnection(SPIF_FPGA_ID, i, board_address, chip_coords)
                for i in SPIF_INPUT_FPGA_LINKS]

    def __outgoing_fpga(
            self, board_address: Optional[str],
            chip_coords: Optional[Tuple[int, int]]) -> FPGAConnection:
        """
        Get the outgoing FPGA connection (for commands).
        """
        return FPGAConnection(
            SPIF_FPGA_ID, SPIF_OUTPUT_FPGA_LINK, board_address, chip_coords)

    def __fpga_indices(self, fpga_link_id: int) -> Tuple[int, int]:
        # We use every other odd link, so we can work out the "index" of the
        # link in the list as follows, and we can then split the index into
        # x and y components
        fpga_index = (fpga_link_id - 1) // 2
        fpga_x_index = fpga_index % self.X_PER_ROW
        fpga_y_index = fpga_index // self.X_PER_ROW
        return fpga_x_index, fpga_y_index

    @overrides(Application2DFPGAVertex.get_incoming_slice_for_link)
    def get_incoming_slice_for_link(
            self, link: FPGAConnection, index: int) -> Slice:
        vertex_slice = super().get_incoming_slice_for_link(link, index)
        self.__index_by_slice[link.fpga_link_id, vertex_slice] = index
        return vertex_slice

    @overrides(Application2DFPGAVertex.get_machine_fixed_key_and_mask)
    def get_machine_fixed_key_and_mask(
            self, machine_vertex: MachineVertex,
            partition_id: str) -> BaseKeyAndMask:
        assert isinstance(machine_vertex, MachineFPGAVertex)
        fpga_link_id = machine_vertex.fpga_link_id
        vertex_slice = machine_vertex.vertex_slice
        index = self.__index_by_slice[fpga_link_id, vertex_slice]
        key_and_mask = self._get_key_and_mask(self.__base_key, index)

        fpga_x, fpga_y = self.__fpga_indices(fpga_link_id)

        # Build the key from the components
        fpga_key = key_and_mask.key + (
            (fpga_y << self._source_y_shift) +
            (fpga_x << self._source_x_shift))
        fpga_mask = key_and_mask.mask | self.__spif_mask
        return BaseKeyAndMask(fpga_key, fpga_mask)

    @overrides(Application2DFPGAVertex.get_fixed_key_and_mask)
    def get_fixed_key_and_mask(self, partition_id: str) -> BaseKeyAndMask:
        n_key_bits = BITS_IN_KEY - self._key_shift
        key_mask = ((1 << n_key_bits) - 1) << self._key_shift
        return BaseKeyAndMask(self.__base_key << self._key_shift, key_mask)

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.start_resume_commands)
    def start_resume_commands(self) -> Iterable[MultiCastCommand]:
        # Make sure everything has stopped
        commands = [SpiNNFPGARegister.STOP.cmd()]

        # Clear the counters
        commands.append(SPIFRegister.OUT_PERIPH_PKT_CNT.cmd(0))
        commands.append(SPIFRegister.CONFIG_PKT_CNT.cmd(0))
        commands.append(SPIFRegister.DROPPED_PKT_CNT.cmd(0))
        commands.append(SPIFRegister.IN_PERIPH_PKT_CNT.cmd(0))

        # Configure the creation of packets from fields to keys using the
        # "standard" input to SPIF (X | P | Y) and convert to (Y | X)
        commands.extend([
            set_field_mask(self.__pipe, 0, self.__input_x_mask),
            set_field_shift(self.__pipe, 0, self.__input_x_shift),
            set_field_limit(self.__pipe, 0,
                            (self.width - 1) << self._source_x_shift),
            set_field_mask(self.__pipe, 1, self.__input_y_mask),
            set_field_shift(self.__pipe, 1, self.__input_y_shift),
            set_field_limit(self.__pipe, 1,
                            (self.height - 1) << self._source_y_shift),
            # These are unused but set them to be sure
            set_field_mask(self.__pipe, 2, 0),
            set_field_shift(self.__pipe, 2, 0),
            set_field_limit(self.__pipe, 2, 0),
            set_field_mask(self.__pipe, 3, 0),
            set_field_shift(self.__pipe, 3, 0),
            set_field_limit(self.__pipe, 3, 0)
        ])

        # Don't filter
        commands.extend([
            set_filter_mask(self.__pipe, i, 0) for i in range(N_FILTERS)
        ])
        commands.extend([
            set_filter_value(self.__pipe, i, 1) for i in range(N_FILTERS)
        ])

        # Configure the output routing key
        commands.append(set_mapper_key(
            self.__pipe, self.__base_key << self._key_shift))

        # Configure the links to send packets to the 8 FPGAs using the
        # lower bits
        commands.extend(
            set_input_key(self.__pipe, i, self.__spif_key(15 - (i * 2)))
            for i in range(8))
        commands.extend(
            set_input_mask(self.__pipe, i, self.__spif_mask)
            for i in range(8))
        commands.extend(
            set_input_route(self.__pipe, i, i)
            for i in range(8))

        # Send the start signal
        commands.append(SpiNNFPGARegister.START.cmd())

        return commands

    def __spif_key(self, fpga_link_id: int) -> int:
        x, y = self.__fpga_indices(fpga_link_id)
        return ((self.__base_key << self._key_shift) +
                (x << self._source_x_shift) +
                (y << self._source_y_shift))

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.pause_stop_commands)
    def pause_stop_commands(self) -> Iterable[MultiCastCommand]:
        # Send the stop signal
        yield SpiNNFPGARegister.STOP.cmd()

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.timed_commands)
    def timed_commands(self) -> List[MultiCastCommand]:
        return []

    @overrides(PopulationApplicationVertex.get_atom_key_map)
    def get_atom_key_map(
            self, pre_vertex: MachineVertex, partition_id: str,
            routing_info: RoutingInfo) -> Iterable[Tuple[int, int]]:
        # Work out which machine vertex
        x_start, y_start = pre_vertex.vertex_slice.start
        key_and_mask = self.get_machine_fixed_key_and_mask(
            pre_vertex, partition_id)
        x_end = x_start + self.sub_width
        y_end = y_start + self.sub_height
        key_x = (key_and_mask.key >> self._source_x_shift) & self.X_MASK
        key_y = (key_and_mask.key >> self._source_y_shift) & self.Y_MASK
        neuron_id = (pre_vertex.vertex_slice.lo_atom +
                     (key_y * self.X_PER_ROW) + key_x)
        for x in range(x_start, x_end, self.X_MASK + 1):
            for y in range(y_start, y_end, self.Y_MASK + 1):
                key = (key_and_mask.key | (x << self._source_x_shift) |
                       (y << self._source_y_shift))
                yield (neuron_id, key)
                neuron_id += self.X_PER_ROW



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.require_subclass import require_subclass
from pacman.model.graphs.machine import MachineVertex


@require_subclass(MachineVertex)
class AbstractMaxSpikes(object, metaclass=AbstractBase):
    """
    Indicates a class (a
    :py:class:`~pacman.model.graphs.machine.MachineVertex`)
    that can describe the maximum rate that it sends spikes.

    The :py:class:`~.SynapticManager` assumes that all machine vertexes
    share the same synapse_information will have the same rates.
    """

    __slots__ = ()

    @abstractmethod
    def max_spikes_per_ts(self) -> float:
        """
        :returns: The maximum expected number of spikes per timestep.
        """
        raise NotImplementedError

    @abstractmethod
    def max_spikes_per_second(self) -> float:
        """
        :returns: The maximum expected number of spikes per second.
        """
        raise NotImplementedError



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides
from spinn_utilities.ranged import RangeDictionary

from spinn_front_end_common.interface.ds import DataType

from spynnaker.pyNN.models.neuron.implementations import ModelParameter
from spynnaker.pyNN.utilities.struct import Struct
from spynnaker.pyNN.data import SpynnakerDataView

from .abstract_additional_input import AbstractAdditionalInput

I_ALPHA = "i_alpha"
I_CA2 = "i_ca2"
TAU_CA2 = "tau_ca2"
TIME_STEP = "time_step"


class AdditionalInputCa2Adaptive(AbstractAdditionalInput):
    """
    The additional model parameters for a leaky integrate and fire model.
    """
    __slots__ = (
        "__tau_ca2",
        "__i_ca2",
        "__i_alpha")

    def __init__(self, tau_ca2: ModelParameter, i_ca2: ModelParameter,
                 i_alpha: ModelParameter):
        r"""
        :param tau_ca2: :math:`\tau_{\mathrm{Ca}^{+2}}`
        :param i_ca2: :math:`I_{\mathrm{Ca}^{+2}}`
        :param i_alpha: :math:`I_{\alpha}`
        """
        super().__init__(
            [Struct([
                 (DataType.S1615, TAU_CA2),
                 (DataType.S1615, I_CA2),
                 (DataType.S1615, I_ALPHA),
                 (DataType.S1615, TIME_STEP)])],
            {I_ALPHA: "nA", I_CA2: "nA", TAU_CA2: "ms"})
        self.__tau_ca2 = tau_ca2
        self.__i_ca2 = i_ca2
        self.__i_alpha = i_alpha

    @overrides(AbstractAdditionalInput.add_parameters)
    def add_parameters(self, parameters: RangeDictionary[float]) -> None:
        parameters[TAU_CA2] = self._convert(self.__tau_ca2)
        parameters[I_ALPHA] = self._convert(self.__i_alpha)
        parameters[TIME_STEP] = SpynnakerDataView.get_simulation_time_step_ms()

    @overrides(AbstractAdditionalInput.add_state_variables)
    def add_state_variables(
            self, state_variables: RangeDictionary[float]) -> None:
        state_variables[I_CA2] = self._convert(self.__i_ca2)

    @property
    def tau_ca2(self) -> ModelParameter:
        r"""
        Settable model parameter: :math:`\tau_{\mathrm{Ca}^{+2}}`
        """
        return self.__tau_ca2

    @property
    def i_ca2(self) -> ModelParameter:
        r"""
        Settable model parameter: :math:`I_{\mathrm{Ca}^{+2}}`
        """
        return self.__i_ca2

    @property
    def i_alpha(self) -> ModelParameter:
        r"""
        Settable model parameter: :math:`I_{\alpha}`
        """
        return self.__i_alpha



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from spinn_utilities.typing.coords import XY
from pacman.model.graphs.application import (
    ApplicationFPGAVertex, FPGAConnection)
from spynnaker.pyNN.models.common import PopulationApplicationVertex


class ArbitraryFPGADevice(ApplicationFPGAVertex, PopulationApplicationVertex):
    """
    A device connected to SpiNNaker via one of the on-board FPGAs.
    """

    __slots__ = ()

    def __init__(
            self, n_neurons: int, fpga_link_id: int, fpga_id: int,
            board_address: Optional[str] = None,
            chip_coords: Optional[XY] = None, label: Optional[str] = None):
        """
        :param n_neurons: Number of neurons
        :param fpga_link_id: The ID of the link of the FPGA (0-15)
        :param fpga_id: The ID of the FPGA on the board (0, 1 or 2)
        :param board_address: The IP address of the board with the FPGA,
            or None for the default board or if using chip_coords
        :param chip_coords: The coordinates of the chip connected to the FPGA,
            or None for the default board or if using board_address
        :param label: The optional name of the vertex.
        """
        conn = FPGAConnection(
            fpga_id, fpga_link_id, board_address, chip_coords)
        super().__init__(n_neurons, [conn], conn, label)



# Copyright (c) 2024 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Sequence, Optional, TYPE_CHECKING

import numpy
from numpy import int32, uint32
from numpy.typing import NDArray

from spinn_utilities.overrides import overrides

from pacman.model.graphs.common import Slice

from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from .abstract_connector import AbstractConnector
from .abstract_generate_connector_on_machine import (
    AbstractGenerateConnectorOnMachine, ConnectorIDs)
from .abstract_generate_connector_on_host import (
    AbstractGenerateConnectorOnHost)

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import SynapseInformation
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge)


class OneToOneOffsetConnector(
        AbstractGenerateConnectorOnMachine,
        AbstractGenerateConnectorOnHost):
    """
    A Connector that connects each pre-neuron to a post-neuron offset by a
    specific amount, positive or negative.  If this goes beyond the start or
    end of the post neurons, it can optionally wrap around.  Additional options
    include a group size, where the offset and wrap is applied repeatedly to
    subsets of neurons.

    In the current implementation it is assumed that the pre- and
    post-populations have the same number of neurons, and that the number of
    neurons is divisible by the group size if specified.  The offset must also
    be smaller than the group size or the number of neurons.
    """

    __slots__ = ("__n_neurons_per_group", "__offset", "__wrap")

    def __init__(self, offset: int, wrap: bool,
                 n_neurons_per_group: Optional[int] = None,
                 safe: bool = True, verbose: bool = False,
                 callback: None = None):
        """
        :param offset:
            The offset to apply to the pre-neuron index to get the post neuron
            index.  This can be positive or negative.
        :param wrap:
            Whether to wrap around the start or end of the post neurons if the
            post neuron id is out of range.
        :param n_neurons_per_group:
            The number of neurons in each group.
            Must be a positive integer divisor of source.size
        :param safe:
            If ``True``, check that weights and delays have valid values.
            If ``False``, this check is skipped.
        :param verbose:
            Whether to output extra information about the connectivity to a
            CSV file
        :param callback:
            if given, a callable that display a progress bar on the terminal.

            .. note::
                Not supported by sPyNNaker.
        """
        super().__init__(safe, callback, verbose)
        self.__n_neurons_per_group = n_neurons_per_group
        self.__offset = offset
        self.__wrap = wrap

    def __n_connections(self, synapse_info: SynapseInformation) -> int:
        if self.__wrap:
            # If there is a wrap, there will always be a next connection
            return synapse_info.n_pre_neurons

        n_groups = 1
        if self.__n_neurons_per_group is not None:
            n_groups = synapse_info.n_pre_neurons // self.__n_neurons_per_group

        # If there isn't a wrap, there are always offset less per group
        return synapse_info.n_pre_neurons - (n_groups * abs(self.__offset))

    @overrides(AbstractConnector.get_delay_maximum)
    def get_delay_maximum(self, synapse_info: SynapseInformation) -> float:
        return self._get_delay_maximum(
            synapse_info.delays, self.__n_connections(synapse_info),
            synapse_info)

    @overrides(AbstractConnector.get_delay_minimum)
    def get_delay_minimum(self, synapse_info: SynapseInformation) -> float:
        return self._get_delay_minimum(
            synapse_info.delays, self.__n_connections(synapse_info),
            synapse_info)

    @overrides(AbstractConnector.get_n_connections_from_pre_vertex_maximum)
    def get_n_connections_from_pre_vertex_maximum(
            self, n_post_atoms: int, synapse_info: SynapseInformation,
            min_delay: Optional[float] = None,
            max_delay: Optional[float] = None) -> int:

        # At most each pre-neuron will one post neuron
        return 1

    @overrides(AbstractConnector.get_n_connections_to_post_vertex_maximum)
    def get_n_connections_to_post_vertex_maximum(
            self, synapse_info: SynapseInformation) -> int:
        # At most each post neuron will be targeted by one pre-neuron
        return 1

    @overrides(AbstractConnector.get_weight_maximum)
    def get_weight_maximum(self, synapse_info: SynapseInformation) -> float:
        return self._get_weight_maximum(
            synapse_info.weights, self.__n_connections(synapse_info),
            synapse_info)

    @overrides(AbstractGenerateConnectorOnHost.create_synaptic_block)
    def create_synaptic_block(
            self, post_slices: Sequence[Slice], post_vertex_slice: Slice,
            synapse_type: int, synapse_info: SynapseInformation) -> NDArray:
        group_size = self.__n_neurons_per_group
        if group_size is None:
            group_size = synapse_info.n_pre_neurons

        # pylint: disable=protected-access
        post_lo, post_hi = synapse_info.pre_population._view_range
        pre_lo, pre_hi = synapse_info.pre_population._view_range
        post_start = max(post_vertex_slice.lo_atom, post_lo)
        post_end = min(post_vertex_slice.hi_atom + 1, post_hi + 1)
        post_group, post_value = divmod(post_start, group_size)

        pre_start = pre_lo + (post_group * group_size)
        pre_end = min(pre_start + group_size, pre_hi)

        pres = list()
        posts = list()
        for post in range(post_start, post_end):
            pre = post - self.__offset
            if pre < pre_start:
                if self.__wrap:
                    pre += group_size
                else:
                    continue
            if pre > pre_end:
                if self.__wrap:
                    pre -= group_size
                else:
                    continue
            pres.append(pre)
            posts.append(post)

            post_value += 1
            if post_value == group_size:
                post_value = 0
                pre_start += group_size
                pre_end = min(pre_start + group_size, pre_hi)
                if pre_start >= pre_hi:
                    break

        block = numpy.zeros(len(pres), dtype=self.NUMPY_SYNAPSES_DTYPE)
        block["source"] = pres
        block["target"] = posts
        block["weight"] = self._generate_weights(
            block["source"], block["target"], len(pres), post_vertex_slice,
            synapse_info)
        block["delay"] = self._generate_delays(
            block["source"], block["target"], len(pres), post_vertex_slice,
            synapse_info)
        block["synapse_type"] = synapse_type
        return block

    def __repr__(self) -> str:
        return (f"offsetConnector(offset={self.__offset}, wrap={self.__wrap}, "
                f"n_neurons_per_group={self.__n_neurons_per_group})")

    @property
    @overrides(AbstractGenerateConnectorOnMachine.gen_connector_id)
    def gen_connector_id(self) -> int:
        return ConnectorIDs.ONE_TO_ONE_OFFSET_CONNECTOR.value

    @overrides(AbstractGenerateConnectorOnMachine.gen_connector_params)
    def gen_connector_params(
            self, synapse_info: SynapseInformation) -> NDArray[uint32]:
        n_values = self.__n_neurons_per_group
        if n_values is None:
            n_values = synapse_info.n_pre_neurons
        return numpy.array(
            [int32(self.__offset).view(uint32), int(self.__wrap), n_values],
            dtype=uint32)

    @property
    @overrides(
        AbstractGenerateConnectorOnMachine.gen_connector_params_size_in_bytes)
    def gen_connector_params_size_in_bytes(self) -> int:
        return BYTES_PER_WORD * 3

    @overrides(AbstractConnector.validate_connection)
    def validate_connection(
            self, application_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> None:
        if (synapse_info.pre_population.size !=
                synapse_info.post_population.size):
            raise NotImplementedError(
                "OneToOneOffsetConnector is only designed to be used with "
                "populations that are the same size as each other")
        if self.__n_neurons_per_group is not None:
            if self.__n_neurons_per_group > synapse_info.pre_population.size:
                raise ValueError(
                    "OneToOneOffsetConnector cannot be used with a group size "
                    "larger than the population size")
            if ((synapse_info.post_population.size /
                 self.__n_neurons_per_group) !=
                    (synapse_info.post_population.size //
                     self.__n_neurons_per_group)):
                raise NotImplementedError(
                    "The number of neurons in each population must be "
                    "divisible by the number of neurons per group")

        n_values = self.__n_neurons_per_group
        if n_values is None:
            n_values = synapse_info.n_pre_neurons
        if n_values < abs(self.__offset):
            raise ValueError(
                "The offset must be smaller than the number of neurons")



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Sequence, Optional, TYPE_CHECKING

import numpy
from numpy import uint32
from numpy.typing import NDArray

from spinn_utilities.overrides import overrides

from pacman.model.graphs.common import Slice

from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from .abstract_connector import AbstractConnector
from .abstract_generate_connector_on_machine import (
    AbstractGenerateConnectorOnMachine, ConnectorIDs)
from .abstract_generate_connector_on_host import (
    AbstractGenerateConnectorOnHost)

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import SynapseInformation


class AllToAllConnector(AbstractGenerateConnectorOnMachine,
                        AbstractGenerateConnectorOnHost):
    """
    Connects all cells in the presynaptic population to all cells in
    the postsynaptic population.
    """

    __slots__ = ("__allow_self_connections", )

    def __init__(self, allow_self_connections: bool = True, safe: bool = True,
                 verbose: bool = False, callback: None = None):
        """
        :param allow_self_connections:
            if the connector is used to connect a Population to itself, this
            flag determines whether a neuron is allowed to connect to itself,
            or only to other neurons in the Population.
        :param safe:
            If ``True``, check that weights and delays have valid values.
            If ``False``, this check is skipped.
        :param verbose:
            Whether to output extra information about the connectivity to a
            CSV file
        :param callback:
            if given, a callable that display a progress bar on the terminal.

            .. note::
                Not supported by sPyNNaker.
        """
        super().__init__(safe, callback, verbose)
        self.__allow_self_connections = allow_self_connections

    @overrides(AbstractConnector.get_delay_maximum)
    def get_delay_maximum(self, synapse_info: SynapseInformation) -> float:
        return self._get_delay_maximum(
            synapse_info.delays,
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            synapse_info)

    @overrides(AbstractConnector.get_delay_minimum)
    def get_delay_minimum(self, synapse_info: SynapseInformation) -> float:
        return self._get_delay_minimum(
            synapse_info.delays,
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            synapse_info)

    @overrides(AbstractConnector.get_n_connections_from_pre_vertex_maximum)
    def get_n_connections_from_pre_vertex_maximum(
            self, n_post_atoms: int, synapse_info: SynapseInformation,
            min_delay: Optional[float] = None,
            max_delay: Optional[float] = None) -> int:
        if min_delay is None or max_delay is None:
            return n_post_atoms

        return self._get_n_connections_from_pre_vertex_with_delay_maximum(
            synapse_info.delays,
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            n_post_atoms, min_delay, max_delay, synapse_info)

    @overrides(AbstractConnector.get_n_connections_to_post_vertex_maximum)
    def get_n_connections_to_post_vertex_maximum(
            self, synapse_info: SynapseInformation) -> int:
        return synapse_info.n_pre_neurons

    @overrides(AbstractConnector.get_weight_maximum)
    def get_weight_maximum(self, synapse_info: SynapseInformation) -> float:
        n_conns = synapse_info.n_pre_neurons * synapse_info.n_post_neurons
        return self._get_weight_maximum(
            synapse_info.weights, n_conns, synapse_info)

    @overrides(AbstractGenerateConnectorOnHost.create_synaptic_block)
    def create_synaptic_block(
            self, post_slices: Sequence[Slice], post_vertex_slice: Slice,
            synapse_type: int, synapse_info: SynapseInformation) -> NDArray:
        n_connections = synapse_info.n_pre_neurons * post_vertex_slice.n_atoms
        no_self = (
            not self.__allow_self_connections and
            synapse_info.pre_population == synapse_info.post_population)
        if no_self:
            n_connections -= post_vertex_slice.n_atoms
        block = numpy.zeros(
            n_connections, dtype=AbstractConnector.NUMPY_SYNAPSES_DTYPE)

        if no_self:
            n_atoms = synapse_info.n_pre_neurons
            sources = numpy.where(numpy.diag(
                numpy.repeat(1, n_atoms)) == 0)[0]
            targets = numpy.array([sources[
                ((n_atoms * i) + (n_atoms - 1)) - j]
                for j in range(n_atoms) for i in range(n_atoms - 1)])
        else:
            sources = numpy.repeat(numpy.arange(
                0, synapse_info.n_pre_neurons), post_vertex_slice.n_atoms)
            targets = numpy.tile(
                numpy.arange(0, post_vertex_slice.n_atoms),
                synapse_info.n_pre_neurons)

        block["source"] = sources
        block["target"] = targets

        block["weight"] = self._generate_weights(
            block["source"], block["target"], n_connections, post_vertex_slice,
            synapse_info)
        block["delay"] = self._generate_delays(
            block["source"], block["target"], n_connections, post_vertex_slice,
            synapse_info)
        block["synapse_type"] = synapse_type
        return block

    def __repr__(self) -> str:
        return "AllToAllConnector()"

    @property
    def allow_self_connections(self) -> bool:
        """
        if the connector is used to connect a Population to itself
        """
        return self.__allow_self_connections

    @allow_self_connections.setter
    def allow_self_connections(self, new_value: bool) -> None:
        self.__allow_self_connections = new_value

    @property
    @overrides(AbstractGenerateConnectorOnMachine.gen_connector_id)
    def gen_connector_id(self) -> int:
        return ConnectorIDs.ALL_TO_ALL_CONNECTOR.value

    @overrides(AbstractGenerateConnectorOnMachine.gen_connector_params)
    def gen_connector_params(
            self, synapse_info: SynapseInformation) -> NDArray[uint32]:
        allow_self = (
            self.__allow_self_connections or
            synapse_info.pre_population != synapse_info.post_population)
        return numpy.array([int(allow_self)], dtype=uint32)

    @property
    @overrides(
        AbstractGenerateConnectorOnMachine.gen_connector_params_size_in_bytes)
    def gen_connector_params_size_in_bytes(self) -> int:
        return BYTES_PER_WORD



# Copyright (c) 2020 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List, Sequence, Tuple
from spinn_utilities.overrides import overrides
from pacman.model.graphs.machine import MachineVertex
from pacman.model.graphs.common import Slice
from pacman.model.resources import AbstractSDRAM
from pacman.model.partitioner_splitters import AbstractSplitterCommon
from pacman.exceptions import PacmanConfigurationException
from pacman.model.partitioner_splitters import SplitterFixedLegacy
from pacman.utilities.utility_objs import ChipCounter
from spynnaker.pyNN.models.spike_source import SpikeSourcePoissonVertex
from .abstract_supports_one_to_one_sdram_input import (
    AbstractSupportsOneToOneSDRAMInput)


class SplitterPoissonDelegate(SplitterFixedLegacy[SpikeSourcePoissonVertex]):
    """
    A splitter for Poisson sources that will ignore sources that are
    one-to-one connected to a single Population.
    """

    @property
    def send_over_sdram(self) -> bool:
        """
        Whether this vertex is to be sent using SDRAM.
        """
        # If there is only one outgoing projection, and it is one-to-one
        # connected to the target, and the target knows what to do, leave
        # it to the target
        if len(self.governed_app_vertex.outgoing_projections) != 1:
            return False
        proj = self.governed_app_vertex.outgoing_projections[0]
        # pylint: disable=protected-access
        post_vertex = proj._projection_edge.post_vertex
        if not isinstance(post_vertex.splitter,
                          AbstractSupportsOneToOneSDRAMInput):
            return False
        return post_vertex.splitter.handles_source_vertex(proj)

    @overrides(SplitterFixedLegacy.set_governed_app_vertex)
    def set_governed_app_vertex(
            self, app_vertex: SpikeSourcePoissonVertex) -> None:
        if not isinstance(app_vertex, SpikeSourcePoissonVertex):
            raise PacmanConfigurationException(
                f"The vertex {app_vertex} cannot be supported by the "
                "SplitterPoissonDelegate as the only vertex supported by this "
                "splitter is a SpikeSourcePoissonVertex. Please use the "
                "correct splitter for your vertex and try again.")
        super().set_governed_app_vertex(app_vertex)

    @overrides(SplitterFixedLegacy.create_machine_vertices)
    def create_machine_vertices(self, chip_counter: ChipCounter) -> None:
        # If sending over SDRAM, let the target handle this
        if self.send_over_sdram:
            return

        # If we passed this part, use the super class
        super().create_machine_vertices(chip_counter)

    @overrides(AbstractSplitterCommon.get_in_coming_slices)
    def get_in_coming_slices(self) -> List[Slice]:
        if self.send_over_sdram:
            proj = self.governed_app_vertex.outgoing_projections[0]
            # pylint: disable=protected-access
            post_vertex = proj._projection_edge.post_vertex
            return list(post_vertex.splitter.get_in_coming_slices())
        return super().get_in_coming_slices()

    @overrides(AbstractSplitterCommon.get_out_going_slices)
    def get_out_going_slices(self) -> List[Slice]:
        if self.send_over_sdram:
            proj = self.governed_app_vertex.outgoing_projections[0]
            # pylint: disable=protected-access
            post_vertex = proj._projection_edge.post_vertex
            return list(post_vertex.splitter.get_out_going_slices())
        return super().get_out_going_slices()

    @overrides(AbstractSplitterCommon.get_out_going_vertices)
    def get_out_going_vertices(self, partition_id: str) -> List[MachineVertex]:
        if self.send_over_sdram:
            return []
        return super().get_out_going_vertices(partition_id)

    @overrides(AbstractSplitterCommon.get_same_chip_groups)
    def get_same_chip_groups(self) -> Sequence[
            Tuple[Sequence[MachineVertex], AbstractSDRAM]]:
        if self.send_over_sdram:
            return []
        return super().get_same_chip_groups()



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from enum import IntEnum
import ctypes
from typing import Optional, Sequence, TYPE_CHECKING

from spinn_utilities.overrides import overrides
from spinn_utilities.abstract_base import abstractmethod
from spinn_utilities.config_holder import get_config_int
from pacman.model.resources import AbstractSDRAM
from pacman.model.graphs.machine import (
    SDRAMMachineEdge, SourceSegmentedSDRAMMachinePartition)
from pacman.model.graphs.common import Slice
from spinn_front_end_common.interface.ds import DataSpecificationGenerator
from spinn_front_end_common.interface.provenance import ProvenanceWriter
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD
from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.exceptions import SynapticConfigurationException
from spynnaker.pyNN.models.abstract_models import (
    ReceivesSynapticInputsOverSDRAM, SendsSynapticInputsOverSDRAM)
from .population_machine_common import CommonRegions, PopulationMachineCommon
from .synaptic_matrices import SynapseRegions
from .population_machine_synapses_provenance import SynapseProvenance
if TYPE_CHECKING:
    from .population_vertex import PopulationVertex
    from .population_neurons_machine_vertex import (
        PopulationNeuronsMachineVertex)

# Size of SDRAM params = 1 word for address + 1 word for size
#  + 1 word for time to send
SDRAM_PARAMS_SIZE = 3 * BYTES_PER_WORD

# Size of the Key config params = 1 work for key + 1 word for mask
#  + 1 word for spike mask + 1 word for colour shift
#  + 1 word for self connection Boolean
KEY_CONFIG_SIZE = 5 * BYTES_PER_WORD


class SpikeProcessingFastProvenance(ctypes.LittleEndianStructure):
    """
    Types of provenance and the DataType used to represent each.
    """
    _fields_ = [
        # A count of the times that the synaptic input circular buffers
        # overflowed
        ("n_buffer_overflows", ctypes.c_uint32),
        # The number of DMA transfers done
        ("n_dmas_complete", ctypes.c_uint32),
        # The number of spikes successfully processed
        ("n_spikes_processed", ctypes.c_uint32),
        # The number of rewires performed.
        ("n_rewires", ctypes.c_uint32),
        # The number of packets that were dropped due to being late
        ("n_late_packets", ctypes.c_uint32),
        # The maximum size of the spike input buffer during simulation
        ("max_size_input_buffer", ctypes.c_uint32),
        # The maximum number of spikes in a time step
        ("max_spikes_received", ctypes.c_uint32),
        # The maximum number of spikes processed in a time step
        ("max_spikes_processed", ctypes.c_uint32),
        # The number of times the transfer time over ran
        ("n_transfer_timer_overruns", ctypes.c_uint32),
        # The number of times a time step was skipped entirely
        ("n_skipped_time_steps", ctypes.c_uint32),
        # The maximum overrun of a transfer
        ("max_transfer_timer_overrun", ctypes.c_uint32),
        # The earliest time a spike was received
        ("earliest_receive", ctypes.c_uint32),
        # The latest time a spike was received
        ("latest_receive", ctypes.c_uint32),
        # The maximum overflow of spikes in a time step
        ("max_spikes_overflow", ctypes.c_uint32)
    ]

    N_ITEMS = len(_fields_)


class PopulationSynapsesMachineVertexCommon(
        PopulationMachineCommon,
        SendsSynapticInputsOverSDRAM):
    """
    Common parts of a machine vertex for the synapses of a Population.
    """

    INPUT_BUFFER_FULL_NAME = "Times_the_input_buffer_lost_packets"
    DMA_COMPLETE = "DMA's that were completed"
    SPIKES_PROCESSED = "How many spikes were processed"
    N_REWIRES_NAME = "Number_of_rewires"
    N_LATE_SPIKES_NAME = "Number_of_late_spikes"
    MAX_FILLED_SIZE_OF_INPUT_BUFFER_NAME = "Max_filled_size_input_buffer"
    MAX_SPIKES_RECEIVED = "Max_spikes_received_in_time_step"
    MAX_SPIKES_PROCESSED = "Max_spikes_processed_in_time_step"
    N_TRANSFER_TIMER_OVERRUNS = "Times_the_transfer_did_not_complete_in_time"
    N_SKIPPED_TIME_STEPS = "Times_a_time_step_was_skipped"
    MAX_TRANSFER_TIMER_OVERRUNS = "Max_transfer_overrun"
    EARLIEST_RECEIVE = "Earliest_receive_time"
    LATEST_RECEIVE = "Latest_receive_time"
    MAX_SPIKE_OVERFLOW = "Max_spike_overflow_in_time_step"

    __slots__ = (
        "__sdram_partition",
        "__neuron_vertex",
        "__partition_id")

    class REGIONS(IntEnum):
        """
        Regions for populations.
        """
        SYSTEM = 0
        PROVENANCE_DATA = 1
        PROFILING = 2
        RECORDING = 3
        SYNAPSE_PARAMS = 4
        SYNAPTIC_MATRIX = 5
        POPULATION_TABLE = 6
        SYNAPSE_DYNAMICS = 7
        STRUCTURAL_DYNAMICS = 8
        BIT_FIELD_FILTER = 9
        SDRAM_EDGE_PARAMS = 10
        KEY_REGION = 11
        CONNECTOR_BUILDER = 12

    # Regions for this vertex used by common parts
    COMMON_REGIONS = CommonRegions(
        REGIONS.SYSTEM,
        REGIONS.PROVENANCE_DATA,
        REGIONS.PROFILING,
        REGIONS.RECORDING)

    # Regions for this vertex used by synapse parts
    SYNAPSE_REGIONS = SynapseRegions(
        REGIONS.SYNAPSE_PARAMS,
        REGIONS.POPULATION_TABLE,
        REGIONS.SYNAPTIC_MATRIX,
        REGIONS.SYNAPSE_DYNAMICS,
        REGIONS.STRUCTURAL_DYNAMICS,
        REGIONS.BIT_FIELD_FILTER,
        REGIONS.CONNECTOR_BUILDER)

    _PROFILE_TAG_LABELS = {
        0: "TIMER_SYNAPSES",
        1: "DMA_READ",
        2: "INCOMING_SPIKE",
        3: "PROCESS_FIXED_SYNAPSES",
        4: "PROCESS_PLASTIC_SYNAPSES"}

    def __init__(
            self, sdram: AbstractSDRAM, label: str,
            app_vertex: PopulationVertex, vertex_slice: Slice):
        """
        :param sdram: The SDRAM used by the vertex
        :param label: The label of the vertex
        :param app_vertex: The associated application vertex
        :param vertex_slice: The slice of the population that this implements
        """
        super().__init__(
            label, app_vertex, vertex_slice, sdram, self.COMMON_REGIONS,
            SynapseProvenance.N_ITEMS + SpikeProcessingFastProvenance.N_ITEMS,
            self._PROFILE_TAG_LABELS, app_vertex.synapse_core_binary_file_name)
        self.__sdram_partition: Optional[
            SourceSegmentedSDRAMMachinePartition] = None
        self.__neuron_vertex: Optional[PopulationNeuronsMachineVertex] = None
        self.__partition_id: Optional[str] = None

    @overrides(SendsSynapticInputsOverSDRAM.set_sdram_partition)
    def set_sdram_partition(
            self,
            sdram_partition: SourceSegmentedSDRAMMachinePartition) -> None:
        if self.__sdram_partition is not None:
            raise SynapticConfigurationException(
                "Trying to set SDRAM partition more than once")
        self.__sdram_partition = sdram_partition

    def set_neuron_vertex_and_partition_id(
            self, neuron_vertex: PopulationNeuronsMachineVertex,
            partition_id: str) -> None:
        """
        Set the neuron vertex and partition ID for the case with a
        self-connection.
        """
        self.__neuron_vertex = neuron_vertex
        self.__partition_id = partition_id

    @overrides(PopulationMachineCommon.get_recorded_region_ids)
    def get_recorded_region_ids(self) -> Sequence[int]:
        ids = self._pop_vertex.synapse_recorder.recorded_ids_by_slice(
            self.vertex_slice)
        assert ids is not None
        return ids

    def _write_sdram_edge_spec(self, spec: DataSpecificationGenerator) -> None:
        """
        Write information about SDRAM Edge.

        :param spec: The generator of the specification to write
        """
        assert self.__sdram_partition is not None
        send_size = self.__sdram_partition.get_sdram_size_of_region_for(self)
        base_addr = self.__sdram_partition.get_sdram_base_address_for(self)

        spec.reserve_memory_region(
            region=self.REGIONS.SDRAM_EDGE_PARAMS,
            size=SDRAM_PARAMS_SIZE, label="SDRAM Params")
        spec.switch_write_focus(self.REGIONS.SDRAM_EDGE_PARAMS)

        spec.write_value(base_addr)
        spec.write_value(send_size)
        spec.write_value(get_config_int(
            "Simulation", "transfer_overhead_clocks"))

    def _write_key_spec(self, spec: DataSpecificationGenerator) -> None:
        """
        Write key configuration region.

        :param spec: The generator of the specification to write
        """
        spec.reserve_memory_region(
            region=self.REGIONS.KEY_REGION, size=KEY_CONFIG_SIZE,
            label="Key Config")
        spec.switch_write_focus(self.REGIONS.KEY_REGION)

        if self.__neuron_vertex is None:
            # No Key = make sure it doesn't match; i.e. spike & 0x0 != 0x1
            spec.write_value(1)  # key
            spec.write_value(0)  # mask
            spec.write_value(0)  # inv_mask
            spec.write_value(0)  # n_colour_bits
            spec.write_value(0)  # is_self_projection
        else:
            assert self.__partition_id is not None
            routing_info = SpynnakerDataView.get_routing_infos()
            r_info = routing_info.get_info_from(
                self.__neuron_vertex, self.__partition_id)
            spec.write_value(r_info.key)
            spec.write_value(r_info.mask)
            spec.write_value(~r_info.mask & 0xFFFFFFFF)
            spec.write_value(self._pop_vertex.n_colour_bits)
            spec.write_value(int(self._pop_vertex.self_projection is not None))

    @overrides(SendsSynapticInputsOverSDRAM.sdram_requirement)
    def sdram_requirement(self, sdram_machine_edge: SDRAMMachineEdge) -> int:
        if isinstance(sdram_machine_edge.post_vertex,
                      ReceivesSynapticInputsOverSDRAM):
            return sdram_machine_edge.post_vertex.n_bytes_for_transfer
        raise SynapticConfigurationException(
            f"Unknown post vertex type in edge {sdram_machine_edge}")

    @overrides(PopulationMachineCommon.parse_extra_provenance_items)
    def parse_extra_provenance_items(
            self, label: str, x: int, y: int, p: int,
            provenance_data: Sequence[int]) -> None:
        proc_offset = SynapseProvenance.N_ITEMS
        self._parse_synapse_provenance(
            label, x, y, p, provenance_data[:proc_offset])
        self._parse_spike_processing_fast_provenance(
            label, x, y, p, provenance_data[proc_offset:])

    @abstractmethod
    def _parse_synapse_provenance(
            self, label: str, x: int, y: int, p: int,
            provenance_data: Sequence[int]) -> None:
        """
        Extract and yield synapse provenance.

        :param label: The label of the node
        :param x: x coordinate of the chip where this core
        :param y: y coordinate of the core where this core
        :param p: virtual id of the core
        :param provenance_data: A list of data items to interpret
        """
        raise NotImplementedError

    def _parse_spike_processing_fast_provenance(
            self, label: str, x: int, y: int, p: int,
            provenance_data: Sequence[int]) -> None:
        """
        Extract and yield spike processing provenance.

        :param label: The label of the node
        :param x: x coordinate of the chip where this core
        :param y: y coordinate of the core where this core
        :param p: virtual id of the core
        :param provenance_data: A list of data items to interpret
        """
        prov = SpikeProcessingFastProvenance(*provenance_data)

        with ProvenanceWriter() as db:
            db.insert_core(
                x, y, p, self.INPUT_BUFFER_FULL_NAME,
                prov.n_buffer_overflows)
            if prov.n_buffer_overflows > 0:
                db.insert_report(
                    f"The input buffer for {label} lost packets on "
                    f"{prov.n_buffer_overflows} occasions. This is often a "
                    "sign that the system is running too quickly for the "
                    "number of neurons per core.  "
                    "Please increase the timer_tic or time_scale_factor or "
                    "decrease the number of neurons per core.")

            db.insert_core(x, y, p, self.DMA_COMPLETE, prov.n_dmas_complete)

            db.insert_core(
                x, y, p, self.SPIKES_PROCESSED, prov.n_spikes_processed)

            db.insert_core(
                x, y, p, self.N_REWIRES_NAME, prov.n_rewires)

            db.insert_core(
                x, y, p, self.N_LATE_SPIKES_NAME, prov.n_late_packets)
            if prov.n_late_packets == 0:
                pass
            elif self._pop_vertex.drop_late_spikes:
                db.insert_report(
                    f"On {label}, {prov.n_late_packets} packets (maximum of "
                    f" {prov.max_spikes_overflow} per time step) were dropped "
                    "from the input buffer, because they arrived too late to "
                    "be processed in a given time step. Try increasing the "
                    "time_scale_factor located within the .spynnaker.cfg file "
                    "or in the pynn.setup() method.")
            else:
                db.insert_report(
                    f"On {label}, {prov.n_late_packets} packets (maximum of "
                    f" {prov.max_spikes_overflow} per time step) arrived too "
                    "late to be processed in a given time step. "
                    "Try increasing the time_scale_factor located within the"
                    " .spynnaker.cfg file or in the pynn.setup() method.")

            db.insert_core(
                x, y, p, self.MAX_FILLED_SIZE_OF_INPUT_BUFFER_NAME,
                prov.max_size_input_buffer)

            db.insert_core(
                x, y, p, self.MAX_SPIKES_RECEIVED, prov.max_spikes_received)

            db.insert_core(
                x, y, p, self.MAX_SPIKES_PROCESSED, prov.max_spikes_processed)

            db.insert_core(
                x, y, p, self.N_TRANSFER_TIMER_OVERRUNS,
                prov.n_transfer_timer_overruns)
            if prov.n_transfer_timer_overruns > 0:
                db.insert_report(
                    f"On {label}, the transfer of synaptic inputs to SDRAM "
                    f"did not end before the next timer tick started "
                    f"{prov.n_transfer_timer_overruns} times with a maximum "
                    f"overrun of {prov.max_transfer_timer_overrun}.  "
                    f"Try increasing transfer_overhead_clocks in your "
                    f".spynnaker.cfg file.")

            db.insert_core(
                x, y, p, self.N_SKIPPED_TIME_STEPS, prov.n_skipped_time_steps)
            if prov.n_skipped_time_steps > 0:
                db.insert_report(
                    f"On {label}, synaptic processing did not start on"
                    f" {prov.n_skipped_time_steps} time steps.  "
                    f"Try increasing the time_scale_factor located within the "
                    f".spynnaker.cfg file or in the pynn.setup() method.")

            db.insert_core(
                x, y, p, self.MAX_TRANSFER_TIMER_OVERRUNS,
                prov.max_transfer_timer_overrun)

            db.insert_core(
                x, y, p, self.EARLIEST_RECEIVE, prov.earliest_receive)
            db.insert_core(
                x, y, p, self.LATEST_RECEIVE, prov.latest_receive)
            db.insert_core(
                x, y, p, self.MAX_SPIKE_OVERFLOW, prov.max_spikes_overflow)



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import struct
from pacman.model.resources import (
    AbstractSDRAM, ConstantSDRAM, VariableSDRAM)
from spinn_front_end_common.utilities.constants import (
    BYTES_PER_WORD, BITS_PER_WORD)

_TWO_WORDS = struct.Struct("<II")


class MultiSpikeRecorder(object):
    """
    Support for recording more than one spike per timestep.
    """
    __slots__ = ("__record", )

    def __init__(self) -> None:
        self.__record = False

    @property
    def record(self) -> bool:
        """
        If this recorder is recording or not
        """
        return self.__record

    @record.setter
    def record(self, record: bool) -> None:
        self.__record = bool(record)

    def get_sdram_usage_in_bytes(
            self, n_neurons: int, spikes_per_timestep: float) -> AbstractSDRAM:
        """
        :param n_neurons:
        :param spikes_per_timestep:
        :returns: SDRAm cost for saving the recordings
        """
        if not self.__record:
            return ConstantSDRAM(0)

        out_spike_bytes = (
            int(math.ceil(n_neurons / BITS_PER_WORD)) * BYTES_PER_WORD)
        return VariableSDRAM(0, (2 * BYTES_PER_WORD) + (
            out_spike_bytes * spikes_per_timestep))



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .abstract_accepts_incoming_synapses import AbstractAcceptsIncomingSynapses
from .abstract_has_delay_stages import AbstractHasDelayStages
from .abstract_max_spikes import AbstractMaxSpikes
from .abstract_synapse_expandable import (
    AbstractSynapseExpandable, SYNAPSE_EXPANDER_APLX)
from .coloured_application_vertex import ColouredApplicationVertex
from .sends_synaptic_inputs_over_sdram import SendsSynapticInputsOverSDRAM
from .receives_synaptic_inputs_over_sdram import (
    ReceivesSynapticInputsOverSDRAM)
from .has_synapses import HasSynapses
from .abstract_neuron_expandable import (
    AbstractNeuronExpandable, NEURON_EXPANDER_APLX)
from .supports_structure import SupportsStructure
from .has_shape_key_fields import HasShapeKeyFields

__all__ = ["AbstractAcceptsIncomingSynapses", "AbstractHasDelayStages",
           "AbstractMaxSpikes", "AbstractSynapseExpandable",
           "ColouredApplicationVertex",
           "SYNAPSE_EXPANDER_APLX",
           "SendsSynapticInputsOverSDRAM", "ReceivesSynapticInputsOverSDRAM",
           "HasSynapses", "AbstractNeuronExpandable", "NEURON_EXPANDER_APLX",
           "SupportsStructure", "HasShapeKeyFields"]



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import ctypes
from typing import Sequence
from spinn_front_end_common.interface.provenance import ProvenanceWriter


class SynapseProvenance(ctypes.LittleEndianStructure):
    """
    Provenance items from synapse processing.
    """
    _fields_ = [
        # A count of presynaptic events.
        ("n_pre_synaptic_events", ctypes.c_uint32),
        # A count of synaptic saturations.
        ("n_saturations", ctypes.c_uint32),
        # The number of STDP weight saturations.
        ("n_plastic_saturations", ctypes.c_uint32),
        # The number of searches of the population table that hit nothing
        ("n_ghost_searches", ctypes.c_uint32),
        # The number of bitfields that couldn't fit in DTCM
        ("n_failed_bitfield_reads", ctypes.c_uint32),
        # The number of population table hits on INVALID entries
        ("n_invalid_pop_table_hits", ctypes.c_uint32),
        # The number of spikes that didn't transfer empty rows
        ("n_filtered_by_bitfield", ctypes.c_uint32),
        # The number of synapses skipped due to late spikes
        ("n_skipped_synapses", ctypes.c_uint32),
        # The number of spikes detected as late
        ("n_late_spikes", ctypes.c_uint32),
        # The maximum lateness of a spike
        ("max_late_spike", ctypes.c_uint32)
    ]

    N_ITEMS = len(_fields_)


class PopulationMachineSynapsesProvenance(object):
    """
    Mix-in to add synapse provenance gathering without other synapse things.
    """

    # This MUST stay empty to allow mixing with other things with slots
    __slots__ = ()

    TOTAL_PRE_SYNAPTIC_EVENT_NAME = "Total_pre_synaptic_events"
    SATURATION_COUNT_NAME = "Times_synaptic_weights_have_saturated"
    SATURATED_PLASTIC_WEIGHTS_NAME = (
        "Times_plastic_synaptic_weights_have_saturated")
    GHOST_SEARCHES = "Number of failed pop table searches"
    BIT_FIELDS_NOT_READ = "N bit fields not able to be read into DTCM"
    INVALID_MASTER_POP_HITS = "Invalid Master Pop hits"
    BIT_FIELD_FILTERED_PACKETS = \
        "How many packets were filtered by the bitfield filterer."
    SYNAPSES_SKIPPED = "Skipped synapses"
    LATE_SPIKES = "Late spikes"
    MAX_LATE_SPIKE = "Max late spike"

    def _parse_synapse_provenance(
            self, label: str, x: int, y: int, p: int,
            provenance_data: Sequence[int]) -> None:
        """
        Extract and yield synapse provenance.

        :param label: The label of the node
        :param x: x coordinate of the chip where this core
        :param y: y coordinate of the core where this core
        :param p: virtual id of the core
        :param provenance_data: A list of data items to interpret
        """
        synapse_prov = SynapseProvenance(*provenance_data)

        with ProvenanceWriter() as db:
            db.insert_core(
                x, y, p, self.TOTAL_PRE_SYNAPTIC_EVENT_NAME,
                synapse_prov.n_pre_synaptic_events)

            db.insert_core(
                x, y, p, self.SATURATION_COUNT_NAME,
                synapse_prov.n_saturations)
            if synapse_prov.n_saturations > 0:
                db.insert_report(
                    f"The weights from the synapses for {label} saturated "
                    f"{synapse_prov.n_saturations} times. ")

            db.insert_core(
                x, y, p, self.SATURATED_PLASTIC_WEIGHTS_NAME,
                synapse_prov.n_plastic_saturations)
            if synapse_prov.n_plastic_saturations > 0:
                db.insert_report(
                    f"The weights from the plastic synapses for {label} "
                    f"saturated {synapse_prov.n_plastic_saturations} times. ")

            db.insert_core(
                x, y, p, self.GHOST_SEARCHES, synapse_prov.n_ghost_searches)
            if synapse_prov.n_ghost_searches > 0:
                db.insert_report(
                    f"The number of failed population table searches for "
                    f"{label} was {synapse_prov.n_ghost_searches}. ")

            db.insert_core(
                x, y, p, self.BIT_FIELDS_NOT_READ,
                synapse_prov.n_failed_bitfield_reads)
            if synapse_prov.n_failed_bitfield_reads:
                db.insert_report(
                    f"On {label}, the filter for stopping redundant DMAs "
                    f"couldn't be fully filled in; it failed to read "
                    f"{synapse_prov.n_failed_bitfield_reads} entries. "
                    "Try reducing neurons per core.")

            db.insert_core(
                x, y, p, self.INVALID_MASTER_POP_HITS,
                synapse_prov.n_invalid_pop_table_hits)
            if synapse_prov.n_invalid_pop_table_hits > 0:
                db.insert_report(
                    f"On {label}, there were "
                    f"{synapse_prov.n_invalid_pop_table_hits} keys received "
                    f"that had no master pop entry for them.")

            db.insert_core(
                x, y, p, self.BIT_FIELD_FILTERED_PACKETS,
                synapse_prov.n_filtered_by_bitfield)

            db.insert_core(
                x, y, p, self.SYNAPSES_SKIPPED,
                synapse_prov.n_skipped_synapses)
            if synapse_prov.n_skipped_synapses > 0:
                db.insert_report(
                    f"On {label}, there were {synapse_prov.n_skipped_synapses}"
                    " that were skipped because a spike was received later"
                    " than the delay in the synapse")

            db.insert_core(
                x, y, p, self.LATE_SPIKES, synapse_prov.n_late_spikes)
            db.insert_core(
                x, y, p, self.MAX_LATE_SPIKE, synapse_prov.max_late_spike)
